#define BLOCK_SIZE 3
#define LOCAL_ADDRESS(index) (index) * get_local_size(0) + get_local_id(0)
#define GLOBAL_ADDRESS(index) (index) * get_global_size(0) + get_global_id(0)

__kernel void align(__global	char	sequences[],
					__global	int		starts_at[],
								int		no_of_subjects,
								float	match_value,
								float	mismatch_value,
								float	gap_opening,
								float	gap_extension,
					__global	float	global_a_value[],
					__global	int		global_a_start[],
					__global	float	global_b_value[],
					__global	int		global_b_start[],
					__local		float	a_value[],
					__local		int		a_start[],
					__local		float	c_value[],
					__local		int		c_start[],
					__local		char	char_block[],
					__global	int		result[])
{
	for (int sequence = get_global_id(0); sequence < no_of_subjects;
				sequence += get_global_size(0)) {
		
		int no_of_rows = starts_at[1] + 1;
		int no_of_cols = starts_at[sequence + 2] - starts_at[sequence + 1] + 1;
		
		__global char	*subject = sequences + starts_at[sequence + 1];
		
		float	max_value = -INFINITY;
		int		max_start;
		int		max_end;
		
		//Not-existing columns are calculated but in the result's calculation attention is paid to this
		for (int col_block = 0; col_block < ceil((float)no_of_cols / BLOCK_SIZE);
				 col_block++) {
			
			//Initialize first row
			{
				int address = LOCAL_ADDRESS(0);
				for (int block_iter = 0; block_iter < BLOCK_SIZE; block_iter++) {
					a_value[address] = 0;
					c_value[address] = -INFINITY;
					char_block[address] = subject[col_block * BLOCK_SIZE + block_iter - 1];
					address += get_local_size(0);
				}
				a_value[address] = 0;
			}
			
			for (int row = 1; row < no_of_rows; row++) {
				float	b_value;
				int		b_start;
				float	left_a_value;
				int		left_a_start;
				char	ref_char = sequences[row - 1];//sequences[0] = reference
				int		gl_row_address = GLOBAL_ADDRESS(row);
			
				if (col_block == 0) {//Initialize first column
					left_a_value = gap_opening + row * gap_extension;
					left_a_start = 1;
					b_value = -INFINITY;
				} else {
					//Load earlier calculated values from the global memory
					b_value = global_b_value[gl_row_address];
					b_start = global_b_start[gl_row_address];
					left_a_value = global_a_value[gl_row_address];
					left_a_start = global_a_start[gl_row_address];
				}
				
				//Iterate through the columns; don't calculate the first column
				for (int block_iter = (col_block == 0 ? 1 : 0);
						block_iter < BLOCK_SIZE; block_iter++) {
					float	first_value;
					float	current_a_value;
					int		current_a_start;
					int		lc_iter_address = LOCAL_ADDRESS(block_iter);
					
					//c
					//a is one field bigger to be able to access the upper left a value
					first_value = a_value[lc_iter_address + get_local_size(0)] + gap_opening;
					if (first_value >= c_value[lc_iter_address]) {
						c_value[lc_iter_address] = first_value;
						c_start[lc_iter_address] = a_start[lc_iter_address 
														+ get_local_size(0)];
					}
					c_value[lc_iter_address] += gap_extension;
					
					//a: first part. Needed for being able to save the upper left a cell(lc_iter_address)
					current_a_value = a_value[lc_iter_address];
					current_a_value += ref_char == char_block[lc_iter_address]
									? match_value : mismatch_value;
					current_a_start = a_start[lc_iter_address];
					
					//Initialize start values
					if (row == 1) {
						c_start[lc_iter_address] = current_a_start =
								col_block * BLOCK_SIZE + block_iter;
					}
					
					//save left a
					a_value[lc_iter_address] = left_a_value;
					a_start[lc_iter_address] = left_a_start;
					
					//b
					left_a_value += gap_opening;//left_a isn't used anymore
					if (left_a_value >= b_value) {
						b_value = left_a_value;
						b_start = left_a_start;
					}
					b_value += gap_extension;
					
					//a
					if (b_value > current_a_value) {
						current_a_value = b_value;
						current_a_start = b_start;
					}
					if (c_value[lc_iter_address] > current_a_value) {
						current_a_value = c_value[lc_iter_address];
						current_a_start = c_start[lc_iter_address];
					}
					left_a_value = current_a_value;
					left_a_start = current_a_start;
				}
				//Save current a value because the next block needs it for the calculation of c
				a_value[LOCAL_ADDRESS(BLOCK_SIZE)] = left_a_value;
				a_start[LOCAL_ADDRESS(BLOCK_SIZE)] = left_a_start;
				global_a_value[gl_row_address] = left_a_value;
				global_a_start[gl_row_address] = left_a_start;
				global_b_value[gl_row_address] = b_value;
				global_b_start[gl_row_address] = b_start;			
			}
			
			//calculate maximum
			int start_col = col_block * BLOCK_SIZE;
			for(int iter = 0;
				iter < min(start_col + BLOCK_SIZE, no_of_cols) - start_col; iter++) {
				if (a_value[LOCAL_ADDRESS(iter + 1)] >= max_value) {
					max_value	= a_value[LOCAL_ADDRESS(iter + 1)];
					max_start	= a_start[LOCAL_ADDRESS(iter + 1)];
					max_end		= start_col + iter;
				}
			}

		}
		//save alignment
		result[sequence * 2] = max_start;
		result[sequence * 2 + 1] = max_end;
	}
}

//val(t) = 0, val(c) = 1, val(a) = 2, val(g) = 3, else val = 64
//Be xyz a codon. Then it's amino acid address is = val(x)*16 + val(y)*4 + val(z)
#define USE_LOCAL 1

#if !USE_LOCAL
__constant char base_to_number[]= {64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,  2, 64,  1, 64, 64, 64,  3, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,  0, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64};
#endif

#if USE_LOCAL
__constant char const_number_to_aa[] = 
#else
__constant char number_to_aa[] = 
#endif
{'F', 'F', 'L', 'L', 'S', 'S', 'S', 'S', 'Y', 'Y', '*', '*', 'C', 'C', '*', 'W', 'L', 'L', 'L', 'L', 'P', 'P', 'P', 'P', 'H', 'H', 'Q', 'Q', 'R', 'R', 'R', 'R', 'I', 'I', 'I', 'M', 'T', 'T', 'T', 'T', 'N', 'N', 'K', 'K', 'S', 'S', 'R', 'R', 'V', 'V', 'V', 'V', 'A', 'A', 'A', 'A', 'D', 'D', 'E', 'E', 'G', 'G', 'G', 'G'};

//sequences: All sequences and the reference sequence concatenated
//starts_at[x]: Where the (x+1)th sequence starts
//interval: interval[2x] and inteval[2x+1] describe the xth sequence's alignment
//translated: The translated null-terminated amino acid sequences 

__kernel void translate(
					__global	char	sequences[],
					__global	int		starts_at[],
								int		no_of_subjects,
					__global	int		interval[],
					__global	char	translated[]
				)
{
#if USE_LOCAL
	//Load arrays into local memory for faster reading
	__local char base_to_number[CHAR_MAX];
	__local char number_to_aa[64];
	for (int i = get_local_id(0); i < CHAR_MAX; i += get_local_size(0))	
		base_to_number[i] = 64;
	barrier(CLK_LOCAL_MEM_FENCE);
	
	if (get_local_id(0) == 0) {
		base_to_number['a'] = 2;
		base_to_number['c'] = 1;
		base_to_number['g'] = 3;
		base_to_number['t'] = 0;
	}
	
	for (int i = get_local_id(0); i < 64; i += get_local_size(0))
		number_to_aa[i] = const_number_to_aa[i];
	barrier(CLK_LOCAL_MEM_FENCE);
#endif

	for(int sequence_no = get_group_id(0); sequence_no < no_of_subjects;
				sequence_no += get_num_groups(0)) {
		//Where the sequence starts
		int sequence_start = starts_at[sequence_no + 1];
		//Where the sequence's alignment starts and ends
		int read_start = sequence_start + interval[sequence_no * 2] - 1;
		int read_end = sequence_start + interval[sequence_no * 2 + 1] - 1;
		//Where to write the amino acid sequence
		int write_start = (sequence_start - starts_at[1]) / 3
							+ sequence_no + get_local_id(0);
		
		//Sequence is null-terminated
		if(get_local_id(0) == 0)
			translated[write_start
			+ (read_end - read_start + 1)/3] = '\0';
		//Translate codons
		for(int i = read_start + 3 * get_local_id(0); i + 2 <= read_end;
					i += 3 * get_local_size(0)) {
			int address = 	  base_to_number[sequences[i]] * 16
							+ base_to_number[sequences[i + 1]] * 4
							+ base_to_number[sequences[i + 2]];
			translated[write_start] = address < 64 ? number_to_aa[address] : 'X';
			write_start += get_local_size(0);
		}
	}
}
