.onLoad <- function(libname, pkgname){
	.C("init", paste(c(libname, "/", pkgname), collapse = ''))
	library(Interpol)
	library(sfsmisc)#AsciiToInt
}

#sequences: A vector of characters Represents the amino acid sequences
#forest: A randomForest.formula. On this random forest the sequences will be classified
#descriptor: An integer between 1 and 533. How the amino acids will be encoded
#normalized.length: An integer. To which length the sequences shall be interpolated
#aa.pos, hull.pos: A row represents a position, a col a coodinate(x,y,z). Where the amino acids and hull points are
classify.on.GPU <- function(sequences, forest, descriptor, normalized.length, aa.pos = NULL, hull.pos = NULL){
	#Create a vector with the descriptor information
	#25 Strings instead of one, because all values are needed and AAdescriptor doesn't add NANs
	#A.pos, Y.Pos: Position of the strings respresenting this characters
	A.pos <- AsciiToInt('A') + 1
	Y.pos <- AsciiToInt('Y') + 1#Z is no amino acid and closing NAs are removed by as.character
	#return values for all ASCII-Signs
	CHAR_MAX <- 127
	
	alphabet <- '0'#\0 is not allowed in strings
	alphabet <- c(alphabet, chars8bit(1:CHAR_MAX))#Get all ASCII-signs
	#Initialize descriptor vector
	aa.descriptor <- rep(as.double(NA), CHAR_MAX)
	#SupressWarnings because as.double warns about NA 
	aa.descriptor[A.pos:Y.pos] <- suppressWarnings(as.double(as.character(AAdescriptor(alphabet[A.pos:Y.pos], descriptor))))
	
	rForest <- forest$forest
	if(!is.null(aa.pos)){
		aa.pos <- as.double(t(aa.pos))
		hull.pos <- as.double(t(hull.pos))
	}
	
	.Call("use_GPU_for_classification", 
		  sequences, 
		  aa.descriptor, as.integer(normalized.length),
		  as.integer(rForest$ntree), as.integer(rForest$treemap[,1,]),
		  as.integer(rForest$bestvar), as.double(rForest$xbestsplit), as.integer(rForest$nodepred),
		  aa.pos, hull.pos)
}